@@ -166,6 +166,9 @@ int compare_version_strings (const char *my_version, const char *req_version);
 /* Format a string so that it fits within about TARGET_COLS columns.  */
 char *format_text (const char *text, int target_cols, int max_cols);
 
+/* Substitute environmen variabales in STRING.  */
+char *substitute_envvars (const char *string);
+
 
 /*-- mapstrings.c --*/
 const char *map_static_macro_string (const char *string);
