@@ -130,6 +130,8 @@ struct
                                the integrity of the software at
                                runtime. */
 
+  unsigned int min_rsa_length;   /* Used for compliance checks.  */
+
   strlist_t keyserver;
 
   /* A list of certificate extension OIDs which are ignored so that
@@ -137,6 +139,10 @@ struct
      OID per string.  */
   strlist_t ignored_cert_extensions;
 
+  /* A list of OIDs which will be used to ignore certificates with
+   * sunch an OID during --learn-card.  */
+  strlist_t ignore_cert_with_oid;
+
   enum gnupg_compliance_mode compliance;
 } opt;
 
